%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%% @doc Standard functions on characters. In Erlang, characters are
%% nonnegative integers representing character codes. Currently, the
%% character set is ISO 8859-1 (Latin-1). This might be extended to full
%% Unicode some day.

-module(erl.lang.char).

-export([escape/1, escape/2, from_integer/1, is_ascii/1, is_char/1,
	 is_control/1, is_digit/1, is_letter/1, is_lower/1,
	 is_printable/1, is_space/1, is_upper/1, nocase_equal/2,
	 to_integer/1, to_lower/1, to_upper/1]).

%% For handling ISO 8859-1 (Latin-1) we use the following type
%% information:
%%
%% 000 - 037	NUL - US	control
%% 040 - 057	SPC - /		punctuation
%% 060 - 071	0 - 9		digit
%% 072 - 100	: - @		punctuation
%% 101 - 132	A - Z		uppercase
%% 133 - 140	[ - `		punctuation
%% 141 - 172	a - z		lowercase
%% 173 - 176	{ - ~		punctuation
%% 177		DEL		control
%% 200 - 237			control
%% 240 - 277	NBSP - 	punctuation
%% 300 - 326	 - 		uppercase
%% 327				punctuation
%% 330 - 336	 - 		uppercase
%% 337 - 366	 - 		lowercase
%% 367				punctuation
%% 370 - 377	 - 		lowercase
%%
%% Many punctuation characters region have special meaning. Must watch
%% using '' (\327), looks very similar to 'x' (\170).


%% These are currently identity functions, since characters *are*
%% integers, but they do not accept invalid inputs.

to_integer(C) when is_integer(C), C >= $\000, C =< $\377 -> C.

from_integer(I) when is_integer(I), I >= $\000, I =< $\377 -> I.

%% Currently, characters are defined as 8-bit Latin1 (codes 0-255).

is_char(C) when is_integer(C), C >= $\000, C =< $\377 -> true;
is_char(_) -> false.


%% Character sets

is_ascii(C) when is_integer(C), C >= $\000, C =< $\177 -> true;
is_ascii(_) -> false.


%% Character ranges

is_digit(C) when is_integer(C), C >= $\060, C =< $\071 -> true;
is_digit(_) -> false.

is_letter(C) when is_integer(C) -> is_letter_1(C);
is_letter(_) -> false.

is_letter_1(C) when C >= $\101, C =< $\132 -> true;
is_letter_1(C) when C >= $\141, C =< $\172 -> true;
is_letter_1(C) when C >= $\300, C =< $\377,
		    C =/= $\327, C =/= $\367 -> true;
is_letter_1(_) -> false.

is_upper(C) when is_integer(C) -> is_upper_1(C);
is_upper(_) -> false.

is_upper_1(C) when C >= $\101, C =< $\132 -> true;
is_upper_1(C) when C >= $\300, C =/= $\327, C =< $\336 -> true;
is_upper_1(_) -> false.

is_lower(C) when is_integer(C) -> is_lower_1(C);
is_lower(_) -> false.

is_lower_1(C) when C >= $\141, C =< $\172 -> true;
is_lower_1(C) when C >= $\337, C =/= $\367, C =< $\377 -> true;
is_lower_1(_) -> false.

is_space($\n) -> true;
is_space($\r) -> true;
is_space($\s) -> true;
is_space($\t) -> true;
is_space(_) -> false.

is_control(C) when is_integer(C) -> is_control_1(C);
is_control(_) -> false.

is_control_1(C) when C >= $\000, C =< $\037 -> true;
is_control_1(C) when C >= $\177, C =< $\237 -> true;
is_control_1(_) -> false.


%% Printable characters

is_printable(C) when is_integer(C) -> is_printable_1(C);
is_printable(_) -> false.

is_printable_1(C) when C >= $\040, C =< $\176 -> true;
is_printable_1(C) when C >= $\240, C =< $\377 -> true;
is_printable_1(C) when C =:= $\b -> true;
is_printable_1(C) when C =:= $\e -> true;
is_printable_1(C) when C =:= $\f -> true;
is_printable_1(C) when C =:= $\n -> true;
is_printable_1(C) when C =:= $\r -> true;
is_printable_1(C) when C =:= $\t -> true;
is_printable_1(C) when C =:= $\v -> true;
is_printable_1(_) -> false.


%% Conversion functions

to_lower(C) when is_integer(C) -> to_lower_1(C).

to_lower_1(C) when C >= $\101, C =< $\132 -> C + $\040;
to_lower_1(C) when C >= $\300, C =/= $\327, C =< $\336 -> C + $\040;
to_lower_1(C) -> C.

to_upper(C) when is_integer(C) -> to_upper_1(C).

to_upper_1(C) when C >= $\141, C =< $\172 -> C - $\040;
to_upper_1(C) when C >= $\340, C =/= $\367, C =< $\376 -> C - $\040;
to_upper_1(C) -> C.


%% Tests

nocase_equal(C1, C2) when is_integer(C1), is_integer(C2) ->
    to_lower_1(C1) =:= to_lower_1(C2);
nocase_equal(_, _) -> false.


%% Character escaping

escape(C) ->
    escape(C, []).

escape(C, Cs) when C >= $\040, C =< $\176 -> [C | Cs];
escape(C, Cs) when C >= $\240, C =< $\377 -> [C | Cs];
escape($\\, Cs) -> [$\\, $\\ | Cs];	% Backslash
escape($\b, Cs) -> [$\\, $b | Cs];	%\b = BS
escape($\d, Cs) -> [$\\, $d | Cs];	%\d = DEL
escape($\e, Cs) -> [$\\, $e | Cs];	%\e = ESC
escape($\f, Cs) -> [$\\, $f | Cs];	%\f = FF
escape($\n, Cs) -> [$\\, $n | Cs];	%\n = LF
escape($\r, Cs) -> [$\\, $r | Cs];	%\r = CR
escape($\t, Cs) -> [$\\, $t | Cs];	%\t = TAB
escape($\v, Cs) -> [$\\, $v | Cs];	%\v = VT
escape(C, Cs) when C >= 0, C =< $\377-> % Octal control characters
    C1 = (C bsr 6) + $0,
    C2 = ((C bsr 3) band 7) + $0,
    C3 = (C band 7) + $0,
    [$\\, C1, C2, C3 | Cs];
escape(C, Cs) -> [C | Cs].	% assume Unicode character
