%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%% @doc Standard functions for calendar dates and time-of-day.

-module(erl.util.date).

-export([add_days/2,
	 day_of_week/0,
	 day_of_week/1,
	 days_in_month/0,
	 days_in_month/1,
	 days_in_month/2,
	 epoch_to_local/1,
	 epoch_to_utc/1,
	 is_leap_year/1,
	 is_valid/1,
	 is_valid_time/1,
	 local/0,
	 local_offset/0,
	 local_to_epoch/1,
	 local_to_utc/1,
	 roll_days/2,
	 second_of_day/1,
	 seconds_to_days/1,
	 seconds_to_time/1,
	 time_of_day/0,
	 today/0,
	 utc/0,
	 utc_to_epoch/1,
	 utc_to_local/1]).


%% Dates are `{{Year, Month, Day}, {Hour, Minute, Second}}'. The
%% Gregorian calendar is used, extended backwards to year 0. Dates can
%% be compared using normal operators ==, /=, >, >=, etc. Arithmetic is
%% done by converting to time measures and back.

%% Terminology: a "day" value is `{Year, Month, Day}" and a "time

-define(SECONDS_PER_MINUTE, 60).
-define(SECONDS_PER_HOUR, 3600).
-define(SECONDS_PER_DAY, 86400).
-define(DAYS_PER_YEAR, 365).
-define(DAYS_PER_LEAP_YEAR, 366).
-define(DAYS_PER_4YEARS, 1461).
-define(DAYS_PER_100YEARS, 36524).
-define(DAYS_PER_400YEARS, 146097).
-define(DAYS_FROM_0_TO_1970, 719528).
-define(MEGA, 1000000).

%% All dates are according the the Gregorian calendar. In this module
%% the Gregorian calendar is extended back to year 0 for convenience.
%%
%% A year Y is a leap year if and only if either
%%
%%        (1)    Y is divisible by 4, but not by 100, or
%%        (2)    Y is divisible by 400.
%%
%% Hence, e.g. 1996 is a leap year, 1900 is not, but 2000 is.
%% 
%% TODO: implement Julian/Gregorian shift.
%%
%% Before the Gregorian calendar, the Julian calendar was used
%% (consistently from March 1, 4AD). The last Julian day was October 4,
%% 1582; the following day, the first Gregorian day, was October 15. The
%% only difference between the systems is that the Gregorian calendar
%% omits century years that are not divisible by 400.

%%
%% EXPORTS
%%

%% @doc Returns the current UTC date.

utc() -> epoch_to_utc(erl.util.time:now()).


%% @doc Returns the current local date.

local() -> epoch_to_local(erl.util.time:now()).


%% @doc Returns the day part of the current local date.

today() ->
    {D, _} = local(),
    D.


%% @doc Returns the time part of the current local date.

time_of_day() ->
    {_, T} = local(),
    T.


%% @doc Converts local date to UTC.

local_to_utc(X) ->
    erl.util.date:local_to_utc(X).	% Built-in function


%% @doc Converts UTC date to local date.

utc_to_local(X) ->
    erl.util.date:utc_to_local(X).	% Built-in function


%% add_days(Days, Date) -> NewDate
%% add_days(Days, {Date, Time}) -> {NewDate, Time}
%% @doc Adds a number of days to a date.

add_days(Days, {Date, Time}) ->
    {add_days(Days, Date), Time};
add_days(Days, Date) when is_integer(Days) ->
    from_days(to_days(Date) + Days).

%% roll_days(Days, Date) -> NewDate
%% roll_days(Days, {Date, Time}) -> {NewDate, Time}
%% @doc Adds a number of days to the day of the month, modulo the total
%% number of days that month. Month and year are not changed.

roll_days(Days, {Date, Time}) ->
    {roll_days(Days, Date), Time};
roll_days(Days, {Y, M, D}) when is_integer(Days) ->
    N = days_in_month(Y, M),
    D0 = (D - 1 + Days) rem N,
    D1 = if D0 >= 0 -> 1 + D0;
	    true -> N + 1 + D0
	 end,
    {Y, M, D1}.


%% @doc Returns the offset for the local time zone, relative to UTC. The
%% returned value is a generic time measure.

local_offset() ->
    U = utc(),
    L = utc_to_local(U),
    time:subtract(utc_to_epoch(L), utc_to_epoch(U)).


%% @doc Returns the day of the week (1-7) for the current local date.

day_of_week() ->
    day_of_week(today()).


%% day_of_week(Date) -> Integer
%% day_of_week({Date, Time}) -> Integer
%%
%% Returns: 1 | .. | 7. Monday = 1, Tuesday = 2, ..., Sunday = 7.

%% @doc Returns the day of the week (1-7) for a particular date.

day_of_week({Date, _}) ->
    day_of_week(Date);
day_of_week(Date) ->
    (to_days(Date) + 5) rem 7 + 1.


%% @spec is_leap_year(Year::integer) -> bool()
%% @doc Returns <code>true</code> if the given year is a leap year,
%% otherwise <code>false</code>.

is_leap_year(Y) when is_integer(Y) ->
    is_leap_year1(Y).

is_leap_year1(Year) when Year rem 4 =:= 0, Year rem 100 =/= 0 ->
    true;
is_leap_year1(Year) when Year rem 400 =:= 0 ->
    true;
is_leap_year1(_) -> false.


%% @doc Returns the number of days in the current month.

days_in_month() ->
    days_in_month(today()).


%% @doc Returns the number of days in the month of a particular date.

days_in_month({D, _}) ->
    days_in_month(D);
days_in_month({Y, M, _}) ->
    days_in_month(Y, M).


%% @spec days_in_month(Year::integer(), Month::integer()) -> integer()
%%
%% @doc Returns the number of days in a month a particular year.

days_in_month(Y, M) when is_integer(Y) ->
    days_in_month_1(Y, M).

days_in_month_1(Y, 2) ->
    case is_leap_year(Y) of
	true -> 29;
	_    -> 28
    end;
days_in_month_1(_, 4) -> 30;
days_in_month_1(_, 6) -> 30;
days_in_month_1(_, 9) -> 30;
days_in_month_1(_,11) -> 30;
days_in_month_1(_, M) when is_integer(M), M >= 1, M =< 12 ->
    31.


%% Args: Date = {{Y, M, D}, {H, M, S}}.
%% Returns: {MegaSec, Sec, MilliSec}.
%%
%% @doc Convert from a UTC date to an epoch-relative time measure.
 
utc_to_epoch({Date,Time}) ->
    S = second_of_day(Time) +
	?SECONDS_PER_DAY * (to_days(Date) - ?DAYS_FROM_0_TO_1970),
    {S div ?MEGA, S rem ?MEGA, 0}.

   
%% Args: Date = {{Y, M, D}, {H, M, S}}.
%% Returns: {MegaSec, Sec, MilliSec}.
%%
%% @doc Convert from a local date to an epoch-relative time measure.

local_to_epoch(Date) ->
    utc_to_epoch(local_to_utc(Date)).

   
%% Args: Time = {MegaSec, Sec, MilliSec},
%% Returns: {{Y, M, D}, {H, Min, S}}.
%%
%% @doc Convert from an epoch-relative time measure to a UTC date.

epoch_to_utc({MSec, Sec, _}) ->
    Sec1 = MSec * ?MEGA + Sec,
    Days = ?DAYS_FROM_0_TO_1970 + Sec1 div ?SECONDS_PER_DAY,
    Rest = Sec1 rem ?SECONDS_PER_DAY,
    {from_days(Days), seconds_to_time(Rest)}.


%% Args: Time = {MegaSec, Sec, MilliSec},
%% Returns: {{Y, M, D}, {H, Min, S}}.
%%
%% @doc Convert from an epoch-relative time measure to a local date.

epoch_to_local(Time) ->
    utc_to_local(epoch_to_utc(Time)).


%% @spec seconds_to_days(Seconds::integer()) -> integer()
%%
%% @doc Converts seconds from 00:00:00 to whole days, also handling
%% negative values. For example, [0..86399] maps to 0, [86400..172799]
%% to 1, etc., and [-1..-86400] to -1, [-86401..-172800] to -2, etc.

seconds_to_days(Secs) when is_integer(Secs) ->
    if Secs >= 0 ->
	    Secs div ?SECONDS_PER_DAY;
       true ->
	    (Secs + 1) div ?SECONDS_PER_DAY - 1
    end.


%% @spec is_valid(Date) -> bool()
%% @doc Returns <code>true</code> if the given value is a valid date,
%% otherwise <code>false</code>.

is_valid({{Y,M,D},{H,Min,S}}) when
  is_integer(Y), is_integer(M), is_integer(D),
  is_integer(H), is_integer(Min), is_integer(S),
  M >= 0, M =< 12, D > 0,
  H >= 0, H < 24, Min >= 0, Min < 60, S >= 0, S < 60 ->
    D =< days_in_month(Y, M);
is_valid(_) -> false.


%% @spec is_valid_time(Time) -> bool()
%% @doc Returns <code>true</code> if the given value is a valid
%% time-of-day, otherwise <code>false</code>.

is_valid_time({H, Min, S}) when
  is_integer(H), is_integer(Min), is_integer(S),
  H >= 0, H < 24, Min >= 0, Min < 60, S >= 0, S < 60 ->
    true;
is_valid_time(_) ->
    false.


%%
%%  LOCAL FUNCTIONS
%%

%% to_days(Date) = Integer
%% to_days({Date, Time}) = Integer
%%
%% Computes the total number of days starting from year 0,
%% January 1st.
%%
%% df/2 catches the case Year<0

to_days({Date, _Time}) ->
    to_days(Date);
to_days({Year, Month, Day}) ->
    dy(Year) + dm(Month) + df(Year, Month) + Day - 1.


%% from_days(Days) = {Year, Month, Day}

from_days(Days) when is_integer(Days) ->
    {Year, DayOfYear} = day_to_year(Days),
    {Month, DayOfMonth} = year_day_to_date(Year, DayOfYear),
    {Year, Month, DayOfMonth}.


%% seconds_to_time(Secs)
%%
%% @doc Returns the time of day corresponding to a number of seconds
%% from 00:00:00. The time count wraps at 23:59:59.

seconds_to_time(Secs) when is_integer(Secs), Secs >= 0 ->
    Secs0 = Secs rem ?SECONDS_PER_DAY,
    Hour = Secs0 div ?SECONDS_PER_HOUR,
    Secs1 = Secs0 rem ?SECONDS_PER_HOUR,
    Minute =  Secs1 div ?SECONDS_PER_MINUTE,
    Second =  Secs1 rem ?SECONDS_PER_MINUTE,
    {Hour, Minute, Second};
seconds_to_time(Secs) when is_integer(Secs), Secs < 0 ->
    seconds_to_time(?SECONDS_PER_DAY + Secs rem ?SECONDS_PER_DAY).


%% second_of_day({Date,Time}) -> Integer
%% second_of_day(Time) -> Integer

%% @doc Returns the number of seconds from 00:00:00 of the given date.

second_of_day({_, Time}) ->
    second_of_day(Time);
second_of_day({H, M, S}) ->
    H * ?SECONDS_PER_HOUR + M * ?SECONDS_PER_MINUTE + S.


%% day_to_year(DayOfEpoch) = {Year, DayOfYear}
%%
%% The idea here is to first guess a year, and then adjust. Although the
%% implementation is recursive, at most 1 or 2 recursive steps are
%% taken.
%% If DayOfEpoch is very large, we need far more than 1 or 2 iterations,
%% since we just subtract a yearful of days at a time until we're there.

day_to_year(DayOfEpoch) when is_integer(DayOfEpoch) ->
    Y0 = DayOfEpoch div ?DAYS_PER_YEAR,
    {Y1, D1} = dty(Y0, DayOfEpoch, dy(Y0)),
    {Y1, DayOfEpoch - D1}.

dty(Y, D1, D2) when D1 < D2 -> 
    dty(Y-1, D1, dy(Y-1));
dty(Y, _D1, D2) ->
    {Y, D2}.


%% year_day_to_date(Year, DayOfYear)  = {Month,  DayOfMonth}
%%
%% Note: 1 is the first day of the month. 

year_day_to_date(Year, DayOfYear) ->
    ExtraDay = case is_leap_year(Year) of
		   true ->
		       1;
		   false ->
		       0
	       end,
    {Month, Day} = year_day_to_date2(ExtraDay, DayOfYear),
    {Month, Day + 1}.
  

%% Note: 0 is the first day of the month 
 
year_day_to_date2(_, Day) when Day < 31 ->
    {1, Day}; 
year_day_to_date2(E, Day) when 31 =< Day, Day < 59 + E ->
    {2, Day - 31}; 
year_day_to_date2(E, Day) when 59 + E =< Day, Day < 90 + E -> 
    {3, Day - (59 + E)}; 
year_day_to_date2(E, Day) when 90 + E =< Day, Day < 120 + E -> 
    {4, Day - (90 + E)};
year_day_to_date2(E, Day) when 120 + E =< Day, Day < 151 + E -> 
    {5, Day - (120 + E)}; 
year_day_to_date2(E, Day) when 151 + E =< Day, Day < 181 + E ->
    {6, Day - (151 + E)};
year_day_to_date2(E, Day) when 181 + E =< Day, Day < 212 + E -> 
    {7, Day - (181 + E)}; 
year_day_to_date2(E, Day) when 212 + E =< Day, Day < 243 + E ->
    {8, Day - (212 + E)};
year_day_to_date2(E, Day) when 243 + E =< Day, Day < 273 + E ->
    {9, Day - (243 + E)};
year_day_to_date2(E, Day) when 273 + E =< Day, Day < 304 + E ->
    {10, Day - (273 + E)};
year_day_to_date2(E, Day) when 304 + E =< Day, Day < 334 + E ->
    {11, Day - (304 + E)};
year_day_to_date2(E, Day) when 334 + E =< Day -> 
    {12, Day - (334 + E)}.


%% dy(Year)
%%
%% Days in previous years.

dy(Y) when Y =< 0 -> 
    0; 
dy(Y) -> 
    X = Y - 1, 
    (X div 4) - (X div 100) + (X div 400) + 
	X*?DAYS_PER_YEAR + ?DAYS_PER_LEAP_YEAR.


%%  dm(Month)
%%
%%  Returns the total number of days in all months
%%  preceeding Month, for an ordinary year.

dm(1) -> 0;    dm(2) ->  31;   dm(3) ->   59;   dm(4) ->  90;
dm(5) -> 120;  dm(6) ->  151;  dm(7) ->  181;   dm(8) -> 212;
dm(9) -> 243;  dm(10) -> 273;  dm(11) -> 304;  dm(12) -> 334.

 
%%  df(Year, Month)
%%
%%  Accounts for an extra day i February if Year is
%%  a leap year, and if Month > 2.

df(_, Month) when Month < 3 ->
    0;
df(Year, _) ->
    case is_leap_year(Year) of
	true -> 1;
	false  -> 0
    end.
